
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <alloca.h>

#include "drmodobj.h"
#include "drmod.h"


static const char *gpstrSym;

extern void eLog (const char *, ...);



/*
** =======================================================================
** =======================================================================
*/
static ObjRelocE 	ArReloc (
									ObjFileT		*ptObjFile,
									ObjSecT			*ptSecTar,
									ObjSecT			*ptSecSym,
									ObjSymT			*ptObjSym,
									Elf32_Rel		*ptRel,
									Elf32_Addr	lAddr)
{
  struct i386_file		*ptIFile 	= (struct i386_file *)		ptObjFile;
  struct i386_symbol	*ptISym  	= (struct i386_symbol *)	ptObjSym;
  Elf32_Addr 					*plLoc 		= (Elf32_Addr *)(ptSecTar->contents + ptRel->r_offset);
  Elf32_Addr 					lOff 			= ptSecTar->header.sh_addr + ptRel->r_offset;
  Elf32_Addr					lGot 			= ptIFile->got ? ptIFile->got->header.sh_addr : 0;

  ObjRelocE	eRC	= eRelocOk;

  switch (ELF32_R_TYPE(ptRel->r_info))
	{
		case R_386_NONE:
		{
			break;
		}

		case R_386_32:
		{
			*plLoc += lAddr;
			break;
		}

		case R_386_PLT32:
		case R_386_PC32:
		{
			*plLoc += lAddr - lOff;
			break;
		}

		case R_386_GLOB_DAT:
		case R_386_JMP_SLOT:
		{
			*plLoc = lAddr;
			break;
		}

		case R_386_RELATIVE:
		{
			*plLoc += ptObjFile->baseaddr;
			break;
		}

		case R_386_GOTPC:
		{
			(lGot == 0) ?
			(eRC = eRelocDanger) :
			(*plLoc += lGot - lOff);
			break;
		}

		case R_386_GOT32:
		{
			if (NULL == ptISym)
			{
				eRC = eRelocDanger;
			}
			else 
			{
				if (! ptISym->gotent.reloc_done)
				{
					ptISym->gotent.reloc_done = 1;
					*(Elf32_Addr *) (ptIFile->got->contents + ptISym->gotent.offset) = lAddr;
				}
				*plLoc += ptISym->gotent.offset;
			}
			break;
		}

		case R_386_GOTOFF:
		{
			(lGot == 0) ?
			(eRC = eRelocDanger) :
			(*plLoc += lAddr - lGot);
			break;
		}

		default:
		{
			eRC = eRelocUnhandle;
			break;
		}
	}
  return eRC;
}



/*
** =======================================================================
** =======================================================================
*/
static void ArFinal (ObjFileT *ptObjFile, Elf32_Addr lBase)
{
  int  i;

  ptObjFile->baseaddr = lBase;

  for (i=0; i<ptObjFile->header.e_shnum; i++)
	{
    ptObjFile->sections[i]->header.sh_addr += lBase;
	}
}


/*
** =======================================================================
** =======================================================================
*/
static void AppendStr (
						const char 		*pstr, 
						char 					**ppstr, 
						ElfW(Word) 		*plStrSize, 
						ElfW(Word) 		*plStrLeft)
{
	int lLen = strlen (pstr) + 1;

	while (lLen > *plStrLeft) 
	{
		*ppstr = realloc (*ppstr, *plStrSize += 4096);

		*plStrLeft += 4096;
	}
	memcpy((char *)*ppstr + *plStrSize - *plStrLeft, pstr, lLen);

	*plStrLeft -= lLen;
}




/*
** =======================================================================
** =======================================================================
*/
static void AppendSym (
						const KallSymT 	*ptSym,
						KallSymT				**pptSym,
						ElfW(Word)			*plSymSize, 
						ElfW(Word)			*plSymLeft)
{
	int lLen = sizeof (*ptSym);

	while (lLen > *plSymLeft) 
	{
		*pptSym = realloc (*pptSym, *plSymSize += 4096);
		*plSymLeft += 4096;
	}
	memcpy ((char *) *pptSym + *plSymSize - *plSymLeft, ptSym, lLen);
	*plSymLeft -= lLen;
}



/*
** =======================================================================
** =======================================================================
*/
static int CompSym (const void *a, const void *b)
{
  KallSymT *c = (KallSymT *) a;
  KallSymT *d = (KallSymT *) b;

  if (c->symbol_addr > d->symbol_addr)
	{
    return 1;
	}
  if (c->symbol_addr < d->symbol_addr)
	{
    return -1;
	}
  return (strcmp (gpstrSym + c->name_off, gpstrSym + d->name_off));
}




/*
** =======================================================================
** =======================================================================
*/
static int ObjLoadLocal (ObjSecT *ptObjSec)
{
  ulong lFlag	= 0;
  ulong lTmp	= 0;

  lFlag = ptObjSec->header.sh_flags;

  if (ptObjSec->name[0] != '.' || strlen(ptObjSec->name) != 10 || 
			strcmp (ptObjSec->name + 5, ".init"))
	{
		lTmp |= 32;
	}
  if (lFlag & SHF_ALLOC) 			lTmp |= 16;
  if (!(lFlag & SHF_WRITE)) 	lTmp |= 8;
  if (lFlag & SHF_EXECINSTR) 	lTmp |= 4;

  if (ptObjSec->header.sh_type != SHT_NOBITS)
	{
		lTmp |= 2;
	}
  return lTmp;
}


/*
** =======================================================================
** =======================================================================
*/
extern ulong ObjHashNcv (const char *pstrSym, ulong lSymLen)
{
  ulong 		lTmp1 = 0;
  ulong 		lTmp2;
  unsigned 	char cChar;

  while (lSymLen > 0)
	{
		cChar = *pstrSym++;
		lTmp1 = (lTmp1 << 4) + cChar;

		if ((lTmp2 = (lTmp1 & 0xf0000000)) != 0)
		{
			lTmp1 ^= lTmp2 >> 24;
			lTmp1 &= ~lTmp2;
		}
		lSymLen--;
	}
  return lTmp1;
}



/*
** =======================================================================
** =======================================================================
*/
extern ulong ObjHash (const char *pstrSym)
{
  return ObjHashNcv (pstrSym, strlen (pstrSym));
}



/*
** =======================================================================
** =======================================================================
*/
extern void	ObjSymCompare (
						ObjFileT 	*ptObjFile,
						int				(*cmp)	(const char *, const char *),
						ulong 		(*hash)	(const char *))
{
  if (NULL != cmp)
	{
    ptObjFile->symbol_cmp = cmp;
	}
  if (NULL != hash)
	{
		ObjSymT *ptTmp[HASH_BUCKETS];
		ObjSymT *ptObjSym	= NULL;
		ObjSymT *ptNext		= NULL;
		int i;

		ptObjFile->symbol_hash = hash;

		memcpy (ptTmp, ptObjFile->symtab, sizeof (ptTmp));
		memset (ptObjFile->symtab, 0, sizeof (ptObjFile->symtab));

		for (i=0; i<HASH_BUCKETS; ++i)
		{
			for (ptObjSym = ptTmp[i]; ptObjSym; ptObjSym = ptNext)
			{
				ulong h = hash (ptObjSym->name) % HASH_BUCKETS;
				ptNext 								= ptObjSym->next;
				ptObjSym->next 				= ptObjFile->symtab[h];
				ptObjFile->symtab[h] 	= ptObjSym;
			}
		}
	}
}




/*
** =======================================================================
** =======================================================================
*/
extern ObjSymT	*ObjAddSymbol (
								ObjFileT		*ptObjFile, 
								const char	*pstrSymName, 
								ulong				lSymIdx,
								int					lInfo, 
								int					lSecIdx, 
								ElfW(Addr)	zValue, 
								ulong				lSize)
{
  ObjSymT 	*ptObjSym	= NULL;
  ulong 		lHash 		= ptObjFile->symbol_hash(pstrSymName) % HASH_BUCKETS;
  int 			lType 		= ELFW(ST_TYPE)(lInfo);
  int 			lBind 		= ELFW(ST_BIND)(lInfo);

  for (ptObjSym = ptObjFile->symtab[lHash]; ptObjSym; ptObjSym = ptObjSym->next)
	{
    if (0 == ptObjFile->symbol_cmp (ptObjSym->name, pstrSymName))
		{
			int lOsecIdx 	= ptObjSym->secidx;
			int lOinfo 		= ptObjSym->info;
			int lOtype 		= ELFW (ST_TYPE) (lOinfo);
			int lObind 		= ELFW (ST_BIND) (lOinfo);

			if (lSecIdx == SHN_UNDEF)
			{
				return ptObjSym;
			}
			else if (lOsecIdx == SHN_UNDEF)
			{
				goto FoundLabel;
			}
			else if (lBind == STB_GLOBAL && lObind == STB_LOCAL)
			{
				ObjSymT *ptObjSymTmp	= NULL;
				ObjSymT **ppObjSym;

				ptObjSymTmp 					= ArNewSym ();
				ptObjSymTmp->next			= ptObjSym->next;
				ptObjSymTmp->ksymidx	= -1;

				for (ppObjSym=&ptObjFile->symtab[lHash]; 
						*ppObjSym != ptObjSym; 
						ppObjSym = &(*ppObjSym)->next)
				{
					continue;
				}
				*ppObjSym = ptObjSym = ptObjSymTmp;
				goto FoundLabel;
			}
			else if (lBind == STB_LOCAL)
			{
				ptObjSym 						= ArNewSym ();
				ptObjSym->next 			= NULL;
				ptObjSym->ksymidx 	= -1;
				ptObjFile->local_symtab[lSymIdx] = ptObjSym;
				goto FoundLabel;
			}
			else if (lBind == STB_WEAK)
			{
				return ptObjSym;
			}
			else if (lObind == STB_WEAK)
			{
				goto FoundLabel;
			}
			else if (lSecIdx == SHN_COMMON && (lOtype == STT_NOTYPE || lOtype == STT_OBJECT))
			{
				return ptObjSym;
			}
			else if (lOsecIdx == SHN_COMMON && (lType == STT_NOTYPE || lType == STT_OBJECT))
			{
				goto FoundLabel;
			}
			else
			{
				if (lSecIdx <= SHN_HIRESERVE)
				{
					eLog("Multiple definition: %s", pstrSymName);
				}
				return ptObjSym;
			}
		}
	}
  ptObjSym 									= ArNewSym();
  ptObjSym->next 						= ptObjFile->symtab[lHash];
  ptObjFile->symtab[lHash] 	= ptObjSym;
  ptObjSym->ksymidx 				= -1;

  if (ELFW(ST_BIND)(lInfo) == STB_LOCAL && lSymIdx != -1) 
	{
    if (lSymIdx >= ptObjFile->local_symtab_size)
		{
      eLog("Local symbol %s with index %ld exceeds local_symtab_size %ld",
			pstrSymName, (long) lSymIdx, (long) ptObjFile->local_symtab_size);
		}
    else
		{
      ptObjFile->local_symtab[lSymIdx] = ptObjSym;
		}
  }

FoundLabel:
  ptObjSym->name 		= pstrSymName;
  ptObjSym->value 	= zValue;
  ptObjSym->size 		= lSize;
  ptObjSym->secidx 	= lSecIdx;
  ptObjSym->info 		= lInfo;
  ptObjSym->r_type 	= 0;

  return ptObjSym;
}





/*
** =======================================================================
** =======================================================================
*/
extern ObjSymT *ObjFindSymbol (ObjFileT *ptObjFile, const char *pstrSymName)
{
  ObjSymT *ptObjSym	= NULL;
  ulong		lHash 		= ptObjFile->symbol_hash(pstrSymName) % HASH_BUCKETS;

  for (ptObjSym = ptObjFile->symtab[lHash]; ptObjSym; ptObjSym = ptObjSym->next)
	{
    if (0 == ptObjFile->symbol_cmp (ptObjSym->name, pstrSymName))
		{
      return ptObjSym;
		}
	}
  return NULL;
}




/*
** =======================================================================
** =======================================================================
*/
extern ElfW(Addr) ObjSymbolValue (ObjFileT *ptObjFile, ObjSymT *ptObjSym)
{
  if (NULL != ptObjSym)
	{
		if (ptObjSym->secidx >= SHN_LORESERVE)
		{
			return ptObjSym->value;
		}
		return ptObjSym->value + ptObjFile->sections[ptObjSym->secidx]->header.sh_addr;
	}
  else
	{
		return 0;
	}
}




/*
** =======================================================================
** =======================================================================
*/
extern ObjSecT *ObjFindSec (ObjFileT *ptObjFile, const char *pstrName)
{
  int i 			= 0;
	int lHdrNum = ptObjFile->header.e_shnum;

  for (i=0; i<lHdrNum; ++i)
	{
    if (0 == strcmp (ptObjFile->sections[i]->name, pstrName))
		{
      return ptObjFile->sections[i];
		}
	}
  return NULL;
}



/*
** =======================================================================
** =======================================================================
*/
extern void ObjInsertSec (ObjFileT *ptObjFile, ObjSecT *ptObjSec)
{
  ObjSecT	**ppObjSec;
  int			lLoad = ObjLoadLocal (ptObjSec);

  for (ppObjSec = ptObjFile->ppStart; 
			*ppObjSec; 
			ppObjSec = &(*ppObjSec)->load_next)
	{
    if (ObjLoadLocal (*ppObjSec) < lLoad)
		{
      break;
		}
	}
  ptObjSec->load_next = *ppObjSec;
  *ppObjSec 					= ptObjSec;
}



/*
** =======================================================================
** =======================================================================
*/
extern ObjSecT	*ObjCreateSec (
								ObjFileT		*ptObjFile, 
								const char	*pstrName,
								ulong				lAlign, 
								ulong				lSize)
{
  ObjSecT *ptObjSec;
  int lIdx = ptObjFile->header.e_shnum++;

  ptObjFile->sections	= realloc (ptObjFile->sections, (lIdx+1) * sizeof (ptObjSec));
  ptObjFile->sections[lIdx] = ptObjSec = ArNewSec ();

  memset (ptObjSec, 0, sizeof(*ptObjSec));
  ptObjSec->header.sh_type			= SHT_PROGBITS;
  ptObjSec->header.sh_flags 		= SHF_WRITE|SHF_ALLOC;
  ptObjSec->header.sh_size 			= lSize;
  ptObjSec->header.sh_addralign = lAlign;
  ptObjSec->name 								= pstrName;
  ptObjSec->idx 								= lIdx;

  if (0 != lSize)
	{
    ptObjSec->contents = malloc(lSize);
	}
  ObjInsertSec (ptObjFile, ptObjSec);

  return ptObjSec;
}



/*
** =======================================================================
** =======================================================================
*/
extern ObjSecT	*ObjCreateSecF (
								ObjFileT		*ptObjFile, 
								const char	*pstrName,
								ulong				lAlign, 
								ulong				lSize)
{
  int lIdx = ptObjFile->header.e_shnum++;
  ObjSecT *ptObjSec;

  ptObjFile->sections = realloc (ptObjFile->sections, (lIdx+1) * sizeof (ptObjSec));
  ptObjFile->sections[lIdx] = ptObjSec = ArNewSec ();

  memset (ptObjSec, 0, sizeof (*ptObjSec));
  ptObjSec->header.sh_type			= SHT_PROGBITS;
  ptObjSec->header.sh_flags 		= SHF_WRITE|SHF_ALLOC;
  ptObjSec->header.sh_size 			= lSize;
  ptObjSec->header.sh_addralign = lAlign;
  ptObjSec->name 								= pstrName;
  ptObjSec->idx 								= lIdx;

  if (0 != lSize)
	{
    ptObjSec->contents = malloc(lSize);
	}
  ptObjSec->load_next 	= ptObjFile->load_order;
  ptObjFile->load_order = ptObjSec;

  if (ptObjFile->ppStart == &ptObjFile->load_order)
	{
    ptObjFile->ppStart = &ptObjSec->load_next;
	}
  return ptObjSec;
}



/*
** =======================================================================
** =======================================================================
*/
extern void *ObjExtendSec (ObjSecT *ptObjSec, ulong lMore)
{
  ulong oldsize = ptObjSec->header.sh_size;

  ptObjSec->contents = realloc (ptObjSec->contents, ptObjSec->header.sh_size += lMore);

  return ptObjSec->contents + oldsize;
}



/*
** =======================================================================
** =======================================================================
*/
extern void ObjStrPt (
						ObjFileT 		*ptObjFile, 
						int 				lSecIdx, 
						ElfW(Addr) 	lOffset, 
						const char 	*pstr)
{
  ObjPtStrT	*ptObjPtStr;
  ObjSecT		*ptObjSec;
  char 			*pstrLoc;
  size_t 		lStrLen = strlen (pstr) + 1;

  ptObjPtStr								= malloc (sizeof (*ptObjPtStr));
  ptObjPtStr->next 					= ptObjFile->string_patches;
  ptObjPtStr->reloc_secidx 	= lSecIdx;
  ptObjPtStr->reloc_offset 	= lOffset;
  ptObjFile->string_patches = ptObjPtStr;

  ptObjSec = ObjFindSec (ptObjFile, ".kstrtab");

  if (NULL == ptObjSec)
	{
		ptObjSec = ObjCreateSec (ptObjFile, ".kstrtab", 1, lStrLen);
		ptObjPtStr->string_offset = 0;
		pstrLoc = ptObjSec->contents;
	}
  else
	{
		ptObjPtStr->string_offset = ptObjSec->header.sh_size;
		pstrLoc = ObjExtendSec (ptObjSec, lStrLen);
	}
  memcpy (pstrLoc, pstr, lStrLen);
}



/*
** =======================================================================
** =======================================================================
*/
extern void ObjSymPt (
						ObjFileT 		*ptObjFile, 
						int 				lSecIdx, 
						ElfW(Addr) 	lOffset, 
						ObjSymT 		*ptObjSym)
{
  ObjPtSymT *ptObjPtSym;

  ptObjPtSym = malloc (sizeof (*ptObjPtSym));

  ptObjPtSym->next 					= ptObjFile->symbol_patches;
  ptObjPtSym->reloc_secidx 	= lSecIdx;
  ptObjPtSym->reloc_offset 	= lOffset;
  ptObjPtSym->sym 					= ptObjSym;
  ptObjFile->symbol_patches = ptObjPtSym;
}


/*
** =======================================================================
** =======================================================================
*/
extern BooleanT ObjUndefV (ObjFileT *ptObjFile)
{
	BooleanT	bSuccess	= TRUE;
  ulong			i;

  for (i=0; i<HASH_BUCKETS; ++i)
	{
		ObjSymT *ptObjSym;

		for (ptObjSym = ptObjFile->symtab[i]; ptObjSym; ptObjSym = ptObjSym->next)
		{
			if (ptObjSym->secidx == SHN_UNDEF)
			{
				if (ELFW (ST_BIND) (ptObjSym->info) == STB_WEAK)
				{
					ptObjSym->secidx = SHN_ABS;
					ptObjSym->value = 0;
				}
				else if (ptObjSym->r_type)
				{
					eLog ("Unresolved symbol: %s", ptObjSym->name);
					bSuccess = FALSE;
				}
			}
		}
	}
  return bSuccess;
}



/*
** =======================================================================
** =======================================================================
*/
extern void ObjAllocCom (ObjFileT *ptObjFile)
{
  struct EntryS
  {
    struct EntryS *next;
    ObjSymT				*ptObjSym;
  } *ptHead = NULL;

  ulong i;

  for (i=0; i<HASH_BUCKETS; ++i)
	{
		ObjSymT *ptObjSym;

		for (ptObjSym = ptObjFile->symtab[i]; ptObjSym; ptObjSym = ptObjSym->next)
		{
			if (ptObjSym->secidx == SHN_COMMON)
			{
				struct EntryS **p, *n;

				for (p = &ptHead; *p; p = &(*p)->next)
				{
					if (ptObjSym->size <= (*p)->ptObjSym->size)
					{
						break;
					}
				}
				n	= alloca (sizeof (*n));
				n->next 		= *p;
				n->ptObjSym = ptObjSym;
				*p = n;
			}
		}
	}

  for (i=1; i<ptObjFile->local_symtab_size; ++i)
	{
		ObjSymT *ptObjSym = ptObjFile->local_symtab[i];

		if (ptObjSym && ptObjSym->secidx == SHN_COMMON)
		{
			struct EntryS **p, *n;

			for (p = &ptHead; *p; p = &(*p)->next)
			{
				if (ptObjSym == (*p)->ptObjSym)
				{
					break;
				}
				else if (ptObjSym->size < (*p)->ptObjSym->size)
				{
					n = alloca (sizeof (*n));
					n->next = *p;
					n->ptObjSym = ptObjSym;
					*p = n;
					break;
				}
			}
		}
	}

  if (ptHead)
	{
		for (i = 0; i < ptObjFile->header.e_shnum; ++i)
		{
			if (ptObjFile->sections[i]->header.sh_type == SHT_NOBITS)
			{
				break;
			}
		}

		if (i == ptObjFile->header.e_shnum)
		{
			ObjSecT *ptObjSec;

			ptObjFile->sections = realloc (ptObjFile->sections, (i + 1) * sizeof (ptObjSec));
			ptObjFile->sections[i] = ptObjSec = ArNewSec ();
			ptObjFile->header.e_shnum = i + 1;

			memset (ptObjSec, 0, sizeof (*ptObjSec));
			ptObjSec->header.sh_type	= SHT_PROGBITS;
			ptObjSec->header.sh_flags = SHF_WRITE | SHF_ALLOC;
			ptObjSec->name 						= ".bss";
			ptObjSec->idx 						= i;
		}

		{
			ElfW (Addr) lShSize	= ptObjFile->sections[i]->header.sh_size;
			ElfW (Addr) lAlign	= ptObjFile->sections[i]->header.sh_addralign;
			struct EntryS *c;

			for (c = ptHead; c; c = c->next)
			{
				ElfW (Addr) align = c->ptObjSym->value;

				if (align > lAlign)
				{
					lAlign = align;
				}
				if (lShSize & (align - 1))
				{
					lShSize = (lShSize | (align - 1)) + 1;
				}
				c->ptObjSym->secidx = i;
				c->ptObjSym->value = lShSize;

				lShSize += c->ptObjSym->size;
			}
			ptObjFile->sections[i]->header.sh_size			= lShSize;
			ptObjFile->sections[i]->header.sh_addralign = lAlign;
		}
	}

  for (i=0; i<ptObjFile->header.e_shnum; ++i)
	{
		ObjSecT *ptObjSec = ptObjFile->sections[i];

		if (ptObjSec->header.sh_type == SHT_NOBITS)
		{
			if (ptObjSec->header.sh_size)
			{
				ptObjSec->contents = 
				memset (malloc (ptObjSec->header.sh_size), 0, ptObjSec->header.sh_size);
			}
			else
			{
				ptObjSec->contents = NULL;
			}
			ptObjSec->header.sh_type = SHT_PROGBITS;
		}
	}
}



/*
** =======================================================================
** =======================================================================
*/
extern ulong ObjLoadSize (ObjFileT *ptObjFile)
{
  ulong			lRC = 0;
  ObjSecT		*ptObjSec;

  for (ptObjSec = ptObjFile->load_order; ptObjSec ; ptObjSec = ptObjSec->load_next)
	{
		ElfW (Addr) lAlign;

		lAlign = ptObjSec->header.sh_addralign;

		if (lAlign && (lRC & (lAlign - 1)))
		{
			lRC = (lRC | (lAlign - 1)) + 1;
		}
		ptObjSec->header.sh_addr = lRC;
		lRC += ptObjSec->header.sh_size;
	}
  return lRC;
}




/*
** =======================================================================
** =======================================================================
*/
extern BooleanT ObjReloc (ObjFileT *ptObjFile, ElfW(Addr) lBase)
{
  BooleanT	bSuccess = TRUE;
  int 			i;

  ArFinal (ptObjFile, lBase);

  for (i=0; i<ptObjFile->header.e_shnum && TRUE == bSuccess; i++)
	{
		ObjSecT	*ptSecRel = ptObjFile->sections[i];

		if (SHT_REL == ptSecRel->header.sh_type)
		{
			ObjSecT 		*ptSecSym	= ptObjFile->sections[ptSecRel->header.sh_link];
			ObjSecT 		*ptSecTar	= ptObjFile->sections[ptSecRel->header.sh_info];
			ObjSecT 		*ptSecStr = ptObjFile->sections[ptSecSym->header.sh_link];
			ElfW(RelM)	*ptRel		= (ElfW (RelM) *) ptSecRel->contents;
			ElfW(RelM)	*ptRelEnd	= ptRel + (ptSecRel->header.sh_size / sizeof (ElfW (RelM)));
			ElfW(Sym)		*ptSymTab	= (ElfW (Sym) *) ptSecSym->contents;
			const char	*ptrTab		= (const char *) ptSecStr->contents;

			for (; ptRel<ptRelEnd && TRUE == bSuccess; ++ptRel)
			{
				ObjSymT			*ptSymInt	= NULL;
				ElfW(Sym)		*ptSymExt	= 0;
				ElfW(Addr)	lAddr 		= 0;
				ulong				lSymIdx		= 0;
				ObjRelocE		eReloc		= eRelocOk;

				if ((lSymIdx = ELFW(R_SYM) (ptRel->r_info)))
				{
					ptSymExt = &ptSymTab[lSymIdx];

					if (ELFW (ST_BIND) (ptSymExt->st_info) == STB_LOCAL)
					{
						ptSymInt = ptObjFile->local_symtab[lSymIdx];
					}
					else
					{
						const char *pstrName;

						if (ptSymExt->st_name)
						{
							pstrName = ptrTab + ptSymExt->st_name;
						}
						else
						{
							pstrName = ptObjFile->sections[ptSymExt->st_shndx]->name;
						}
						ptSymInt = ObjFindSymbol (ptObjFile, pstrName);
					}
					lAddr = ObjSymbolValue (ptObjFile, ptSymInt);
				}

				eReloc = ArReloc (ptObjFile, ptSecTar, ptSecSym, ptSymInt, ptRel, lAddr);

				if (eRelocOk != eReloc)
				{
					eLog("Obj relocation failure");
					bSuccess = FALSE;
				}
			}
		}
	}

  if (TRUE == bSuccess && ptObjFile->string_patches)
	{
		ObjPtStrT		*ptObjStr;
		ObjSecT			*ptSecStr;
		ElfW(Addr)	lAddrBase;

		ptSecStr	= ObjFindSec (ptObjFile, ".kstrtab");
		lAddrBase = ptSecStr->header.sh_addr;

		for (ptObjStr = ptObjFile->string_patches; ptObjStr; ptObjStr = ptObjStr->next)
		{
			ObjSecT *ptSecTar = ptObjFile->sections[ptObjStr->reloc_secidx];
			*(ElfW (Addr) *) (ptSecTar->contents + ptObjStr->reloc_offset) = 
			lAddrBase + ptObjStr->string_offset;
		}
	}

  if (TRUE == bSuccess && ptObjFile->symbol_patches)
	{
		ObjPtSymT *ptPtSym;

		for (ptPtSym = ptObjFile->symbol_patches; ptPtSym; ptPtSym = ptPtSym->next)
		{
			ObjSecT *ptSecTar = ptObjFile->sections[ptPtSym->reloc_secidx];
			*(ElfW (Addr) *) (ptSecTar->contents + ptPtSym->reloc_offset) = 
			ObjSymbolValue (ptObjFile, ptPtSym->sym);
		}
	}
  return bSuccess;
}



/*
** =======================================================================
** =======================================================================
*/
extern void ObjCreateImg (ObjFileT *ptObjFile, char *ptrImg)
{
  ObjSecT			*ptObjSec;
  ElfW(Addr)	lBase = ptObjFile->baseaddr;

  for (ptObjSec = ptObjFile->load_order; ptObjSec ; ptObjSec = ptObjSec->load_next)
	{
		char *pstrSecImg;

		if (ptObjSec->contents == 0)
		{
			continue;
		}
		pstrSecImg = ptrImg + (ptObjSec->header.sh_addr - lBase);
		memcpy (pstrSecImg, ptObjSec->contents, ptObjSec->header.sh_size);
	}
}



/*
** =======================================================================
** =======================================================================
*/
extern ObjFileT *ArNewFile ()
{
  struct i386_file *ptFile = malloc (sizeof (*ptFile));

  ptFile->got = NULL;

  return &ptFile->root;
}



/*
** =======================================================================
** =======================================================================
*/
extern ObjSecT *ArNewSec ()
{
  return malloc (sizeof (ObjSecT));
}



/*
** =======================================================================
** =======================================================================
*/
extern ObjSymT *ArNewSym ()
{
  struct i386_symbol *ptSym = malloc (sizeof (*ptSym));

  memset (&ptSym->gotent, 0, sizeof (ptSym->gotent));

  return &ptSym->root;
}




/*
** =======================================================================
** =======================================================================
*/
extern void ArCreate (ObjFileT *ptObjFile)
{
  struct i386_file	*ptIFile 	= (struct i386_file *) ptObjFile;
  int								lOffset 	= 0;
  int								lNeed 		= 0;
  int 							i					= 0;

  for (i=0; i<ptIFile->root.header.e_shnum; i++)
	{
		ObjSecT 		*ptSecRel;
		ObjSecT 		*ptSecSym;
		ObjSecT 		*ptSecStr;
		Elf32_Sym 	*ptSymTab;
		Elf32_Rel		*ptRelEnd;
		Elf32_Rel		*ptRel;
		const char 	*ptrTab;

		ptSecRel = ptIFile->root.sections[i];

		if (ptSecRel->header.sh_type != SHT_REL)
		{
			continue;
		}
		ptSecSym 	= ptIFile->root.sections[ptSecRel->header.sh_link];
		ptSecStr 	= ptIFile->root.sections[ptSecSym->header.sh_link];
		ptRel 		= (Elf32_Rel *) ptSecRel->contents;
		ptRelEnd 	= ptRel + (ptSecRel->header.sh_size / sizeof (Elf32_Rel));
		ptSymTab 	= (Elf32_Sym *) ptSecSym->contents;
		ptrTab 		= (const char *) ptSecStr->contents;

		for (; ptRel < ptRelEnd; ++ptRel)
		{
			Elf32_Sym 					*ptSymExt;
			struct i386_symbol 	*ptSymInt;
			const char 					*pstrName;

			switch (ELF32_R_TYPE (ptRel->r_info))
			{
				case R_386_GOTPC:
				case R_386_GOTOFF:
				{
					lNeed = 1;
				}
				default:
				{
					continue;
				}
				case R_386_GOT32:
				{
					break;
				}
			}
			ptSymExt = &ptSymTab[ELF32_R_SYM (ptRel->r_info)];

			if (ptSymExt->st_name)
			{
				pstrName = ptrTab + ptSymExt->st_name;
			}
			else
			{
				pstrName = ptObjFile->sections[ptSymExt->st_shndx]->name;
			}
			ptSymInt = (struct i386_symbol *) ObjFindSymbol (&ptIFile->root, pstrName);

			if (! ptSymInt->gotent.offset_done)
			{
				ptSymInt->gotent.offset_done = 1;
				ptSymInt->gotent.offset = lOffset;
				lOffset += 4;
			}
		}
	}

  if (lOffset > 0 || lNeed)
	{
    ptIFile->got = ObjCreateSec (&ptIFile->root, ".got", 4, lOffset);
	}
}



/*
** =======================================================================
** =======================================================================
*/
extern void ObjKallSym (ObjFileT *ptObjFile, ObjFileT **pptObjFile)
{
  ObjFileT			*ptFile;
  ObjSecT				*ptSecI;
  ObjSecT				*ptSecO;
  KallSymT			*ptKallSym = NULL;
  KallSymT			tKallSym;
  KallSymHdrT		*ptSymHdr;
  KallSymSecT		*ptSymSec;
  ElfW(Off)			lSecOff;
  ElfW(Off)			lFileOff;
  ElfW(Word)		lSymSize = 0;
  ElfW(Word)		lSymLeft = 0;
  ElfW(Word)		lStrSize = 0;
  ElfW(Word)		lStrLeft = 0;
  char					*pstr = NULL;
  char					*pstrTmp;
  int						*plSymMap;
  int						i;
  int						lLoad = 0;
  static char		strTab[] = "\000" KALLSYMS_SEC_NAME;

  ptFile = ArNewFile ();
  memset (ptFile, 0, sizeof (*ptFile));

  ptFile->symbol_cmp 	= strcmp;
  ptFile->symbol_hash = ObjHash;
  ptFile->ppStart 		= &ptFile->load_order;

  memcpy (&ptFile->header, &ptObjFile->header, sizeof (ptFile->header));
  ptFile->header.e_type 			= ET_REL;
  ptFile->header.e_entry 			= 0;
  ptFile->header.e_phoff 			= 0;

  lFileOff = sizeof (ptFile->header);

  ptFile->header.e_shoff 			= lFileOff;
  ptFile->header.e_phentsize 	= 0;
  ptFile->header.e_phnum 			= 0;
  ptFile->header.e_shnum 			= KALLSYMS_IDX + 1;
  ptFile->header.e_shstrndx 	= KALLSYMS_IDX - 1;

  lFileOff += ptFile->header.e_shentsize * ptFile->header.e_shnum;
  ptFile->sections = malloc (sizeof (*(ptFile->sections)) * ptFile->header.e_shnum);
  memset (ptFile->sections, 0, sizeof (*(ptFile->sections)) * ptFile->header.e_shnum);
  ptFile->sections[0]	= ptSecO = ArNewSec ();
  memset (ptSecO, 0, sizeof (*ptSecO));
  ptSecO->header.sh_type 	= SHT_NULL;
  ptSecO->header.sh_link 	= SHN_UNDEF;

  ptFile->sections[KALLSYMS_IDX - 1] = ptSecO = ArNewSec ();
  memset (ptSecO, 0, sizeof (*ptSecO));
  ptSecO->name 							= ".strtab";
  ptSecO->header.sh_type 		= SHT_STRTAB;
  ptSecO->header.sh_link 		= SHN_UNDEF;
  ptSecO->header.sh_offset 	= lFileOff;
  ptSecO->header.sh_size 		= sizeof (strTab);
  ptSecO->contents 					= malloc (sizeof (strTab));
  memcpy (ptSecO->contents, strTab, sizeof (strTab));
  lFileOff += ptSecO->header.sh_size;

  ptFile->sections[KALLSYMS_IDX] = ptSecO = ArNewSec ();
  memset (ptSecO, 0, sizeof (*ptSecO));
  ptSecO->name 						= KALLSYMS_SEC_NAME;
  ptSecO->header.sh_name 	= 1;
  ptSecO->header.sh_type 	= SHT_PROGBITS;
  ptSecO->header.sh_flags = SHF_ALLOC;
  ptSecO->header.sh_link 	= SHN_UNDEF;
  ptSecO->header.sh_addralign = sizeof (ElfW (Word));
  lFileOff = (lFileOff + ptSecO->header.sh_addralign - 1) & -(ptSecO->header.sh_addralign);
  ptSecO->header.sh_offset = lFileOff;

  for (i=0; i<ptObjFile->header.e_shnum; ++i)
	{
		if (ptObjFile->sections[i]->header.sh_flags & SHF_ALLOC)
		{
			++lLoad;
		}
	}

  ptSecO->header.sh_size = sizeof (*ptSymHdr) + lLoad * sizeof (*ptSymSec);
  ptSymHdr = (KallSymHdrT *) ptSecO->contents = malloc (ptSecO->header.sh_size);
  memset (ptSecO->contents, 0, ptSecO->header.sh_size);
  ptSymHdr->size = sizeof (*ptSymHdr);
  ptSymHdr->sections = lLoad;
  ptSymHdr->section_off = ptSymHdr->size;
  ptSymHdr->section_size = sizeof (*ptSymSec);
  ptSymHdr->symbol_off = ptSecO->header.sh_size;
  ptSymHdr->symbol_size = sizeof (tKallSym);
  ptSymHdr->start = (ElfW (Addr)) (~0);
  lSecOff 	= 0;
  plSymMap 	= malloc (sizeof (*plSymMap) * ptObjFile->header.e_shnum);

  for (i=0; i<ptObjFile->header.e_shnum; ++i)
	{
		ptSecI = ptObjFile->sections[i];

		if (ptSecI->header.sh_flags & SHF_ALLOC)
		{
			plSymMap[ptSecI->idx] = lSecOff;
			lSecOff += ptSymHdr->section_size;
		}
		else
		{
			plSymMap[ptSecI->idx] = -1;
		}
	}
  ptSymSec = (KallSymSecT *) ((char *) ptSymHdr + ptSymHdr->section_off);

  for (i=0; i<ptObjFile->header.e_shnum; ++i)
	{
		ptSecI = ptObjFile->sections[i];

		if (! (ptSecI->header.sh_flags & SHF_ALLOC))
		{
			continue;
		}
		ptSymSec->start 		= ptSecI->header.sh_addr;
		ptSymSec->size 			= ptSecI->header.sh_size;
		ptSymSec->flags 		= ptSecI->header.sh_flags;
		ptSymSec->name_off 	= lStrSize - lStrLeft;
		AppendStr (ptSecI->name, &pstr, &lStrSize, &lStrLeft);

		if (ptSymSec->start < ptSymHdr->start)
		{
			ptSymHdr->start = ptSymSec->start;
		}
		if (ptSymSec->start + ptSymSec->size > ptSymHdr->end)
		{
			ptSymHdr->end = ptSymSec->start + ptSymSec->size;
		}
		++ptSymSec;
	}

  for (i=0; i<HASH_BUCKETS; ++i)
	{
		ObjSymT *ptObjSym = ptObjFile->symtab[i];

		for (ptObjSym = ptObjFile->symtab[i]; ptObjSym; ptObjSym = ptObjSym->next)
		{
			if (! ptObjSym || ptObjSym->secidx >= ptObjFile->header.e_shnum)
			{
				continue;
			}
			if ((tKallSym.section_off = plSymMap[ptObjSym->secidx]) == -1)
			{
				continue;
			}
			if (strcmp (ptObjSym->name, "gcc2_compiled.") == 0 || 
				 strncmp (ptObjSym->name, "__insmod_", 9) == 0)
			{
				continue;
			}
			tKallSym.symbol_addr = ptObjSym->value;

			if (ptObjFile->header.e_type == ET_REL)
			{
				tKallSym.symbol_addr += ptObjFile->sections[ptObjSym->secidx]->header.sh_addr;
			}
			tKallSym.name_off = lStrSize - lStrLeft;
			AppendSym (&tKallSym, &ptKallSym, &lSymSize, &lSymLeft);
			AppendStr (ptObjSym->name, &pstr, &lStrSize, &lStrLeft);
			++ptSymHdr->symbols;
		}
	}
  free (plSymMap);

  gpstrSym = pstr;
  qsort ((char *) ptKallSym, (unsigned) ptSymHdr->symbols, sizeof (*ptKallSym), CompSym);
  gpstrSym = NULL;
  ptSecO->header.sh_size = ptSymHdr->total_size = 
	ptSymHdr->symbol_off + ptSymHdr->symbols * ptSymHdr->symbol_size + lStrSize - lStrLeft;
  ptSymHdr = (KallSymHdrT *) ptSecO->contents = realloc (ptSymHdr, ptSymHdr->total_size);
  pstrTmp = (char *) ptSymHdr + ptSymHdr->symbol_off;
  memcpy (pstrTmp, ptKallSym, ptSymHdr->symbols * ptSymHdr->symbol_size);
  free (ptKallSym);
  pstrTmp += ptSymHdr->symbols * ptSymHdr->symbol_size;
  ptSymHdr->string_off = pstrTmp - (char *) ptSymHdr;
  memcpy (pstrTmp, pstr, lStrSize - lStrLeft);
  free (pstr);
  *pptObjFile = ptFile;
}
